==> effconnpy-0.1.0.tar.gz: effconnpy-0.1.0/effconnpy/extendedanalyzers.py <==
import numpy as np
import pandas as pd
from typing import Union, List, Optional
import pymc3 as pm
import arviz as az
import networkx as nx
import dowhy
from sklearn.preprocessing import StandardScaler
from causalnex.network import BayesianNetwork
from causalnex.structure import StructureModel
from causalnex.structure.notears import from_pandas
from .causality_analyzer import CausalityAnalyzer

class ExtendedCausalityAnalyzer(CausalityAnalyzer):
    def __init__(self, data: Union[np.ndarray, pd.DataFrame, List[np.ndarray]]):
        """
        Initialize ExtendedCausalityAnalyzer with extended causal inference techniques
        
        Args:
            data: Input time series data
        """
        super().__init__(data)
        
        # Convert to DataFrame for some methods
        self.df = pd.DataFrame(self.data)
    
    def dynamic_bayesian_network(self, max_lag: int = 1, threshold: float = 0.3):
        """
        Construct Dynamic Bayesian Network
        
        Args:
            max_lag: Maximum number of time lags to consider
            threshold: Correlation threshold for edge creation
        
        Returns:
            Bayesian Network structure
        """
        # Create structure model
        sm = StructureModel()
        
        # Iterate through lags
        for lag in range(1, max_lag + 1):
            # Create lagged features
            lagged_df = pd.DataFrame()
            for col in self.df.columns:
                lagged_df[f"{col}_lag_{lag}"] = self.df[col].shift(lag)
            
            # Combine original and lagged data
            combined_df = pd.concat([self.df, lagged_df], axis=1).dropna()
            
            # Learn structure
            structure = from_pandas(combined_df, w_threshold=threshold)
            
            # Add edges to structure model
            for edge in structure.edges():
                sm.add_edge(edge[0], edge[1])
        
        # Create Bayesian Network
        bn = BayesianNetwork(sm)
        
        return bn
    
    def structural_equation_modeling(self):
        """
        Perform Structural Equation Modeling
        
        Returns:
            SEM model results and parameters
        """
        # Standardize data
        scaler = StandardScaler()
        scaled_data = scaler.fit_transform(self.data)
        
        # Use PyMC3 for SEM
        with pm.Model() as sem_model:
            # Latent variables
            latent_vars = {}
            for i in range(self.num_series):
                latent_vars[f'latent_{i}'] = pm.Normal(f'latent_{i}', 0, 1)
            
            # Structural relationships
            observations = {}
            for i in range(self.num_series):
                observations[f'obs_{i}'] = pm.Normal(
                    f'obs_{i}', 
                    mu=sum(latent_vars.values()),  # Simple summed latent effect
                    sigma=1
                )
            
            # Observe the data
            pm.Normal('likelihood', mu=observations, observed=scaled_data)
        
        # Sample from the model
        with sem_model:
            trace = pm.sample(2000, return_inferencedata=True)
        
        return {
            'model': sem_model,
            'trace': trace,
            'summary': az.summary(trace)
        }
    
    def causal_discovery_dowhy(self, treatment_var: int = 0, outcome_var: int = 1):
        """
        Causal Discovery using DoWhy
        
        Args:
            treatment_var: Index of treatment variable
            outcome_var: Index of outcome variable
        
        Returns:
            Causal inference results
        """
        # Prepare data
        data = self.df.copy()
        
        # Create causal graph
        graph = nx.DiGraph()
        graph.add_edges_from([(f'X{treatment_var}', f'X{outcome_var}')])
        
        # DoWhy causal model
        causal_model = dowhy.CausalModel(
            data=data,
            treatment=f'X{treatment_var}',
            outcome=f'X{outcome_var}',
            graph=graph
        )
        
        # Identify causal effect
        identified_estimand = causal_model.identify_effect()
        
        # Estimate causal effect
        estimate = causal_model.estimate_effect(identified_estimand)
        
        return {
            'causal_model': causal_model,
            'identified_estimand': identified_estimand,
            'estimate': estimate
        }
    
    def dynamic_causal_modeling(self):
        """
        Dynamic Causal Modeling (DCM)
        
        Returns:
            DCM model results
        """
        # Basic DCM implementation using PyMC3
        with pm.Model() as dcm_model:
            # State transitions
            x = pm.GaussianRandomWalk('state', sigma=1, shape=self.num_series)
            
            # Observation model
            obs_sigma = pm.HalfNormal('obs_sigma', sigma=1)
            obs = pm.Normal('obs', mu=x, sigma=obs_sigma, observed=self.data)
        
        # Sample from the model
        with dcm_model:
            trace = pm.sample(2000, return_inferencedata=True)
        
        return {
            'model': dcm_model,
            'trace': trace,
            'summary': az.summary(trace)
        }
    
    def causality_test(self, 
                       method: str = 'granger', 
                       lag: Optional[int] = None, 
                       verbose: bool = False, 
                       **kwargs):
        """
        Extended causality test method
        
        Args:
            method: Causality test method
            lag: Number of lags
            verbose: Verbose output
            **kwargs: Additional method-specific parameters
        
        Returns:
            Causality test results
        """
        # Existing methods plus new ones
        methods = {
            'granger': self.granger_causality,
            'dbn': self.dynamic_bayesian_network,
            'sem': self.structural_equation_modeling,
            'dowhy': self.causal_discovery_dowhy,
            'dcm': self.dynamic_causal_modeling
        }
        
        if method.lower() not in methods:
            raise ValueError(f"Method {method} not supported. Choose from {list(methods.keys())}")
        
        return methods[method.lower()](**kwargs)

==> effconnpy-0.1.6.tar.gz: effconnpy-0.1.6/effconnpy/extendedanalyzers.py <==
import numpy as np
import pandas as pd
from typing import Union, List, Optional
import pymc3 as pm
import arviz as az
import networkx as nx
import dowhy
from sklearn.preprocessing import StandardScaler
from scipy.stats import pearsonr
from .causality_analyzer import CausalityAnalyzer

class ExtendedCausalityAnalyzer(CausalityAnalyzer):
    def __init__(self, data: Union[np.ndarray, pd.DataFrame, List[np.ndarray]]):
        """
        Initialize ExtendedCausalityAnalyzer with extended causal inference techniques
        
        Args:
            data: Input time series data
        """
        super().__init__(data)
        
        # Convert to DataFrame for some methods
        self.df = pd.DataFrame(self.data)
    
   def dynamic_bayesian_network(self, max_lag: int = 1, correlation_threshold: float = 0.3):
    """
    Construct Dynamic Bayesian Network using NetworkX
    
    Args:
        max_lag: Maximum number of time lags to consider
        correlation_threshold: Correlation threshold for edge creation
    
    Returns:
        Directed graph representing causal relationships
    """
    # Create directed graph
    dbn = nx.DiGraph()
    
    # Create lagged features
    lagged_df = pd.DataFrame()
    for col in self.df.columns:
        for lag in range(1, max_lag + 1):
            lagged_df[f"{col}_lag_{lag}"] = self.df[col].shift(lag)
    
    # Combine original and lagged data
    combined_df = pd.concat([self.df, lagged_df], axis=1).dropna()
    
    # Compute correlations and create edges
    for col1 in self.df.columns:
        for col2 in self.df.columns:
            if col1 != col2:
                # Check current and lagged correlations
                for lag in range(1, max_lag + 1):
                    corr, p_value = pearsonr(
                        combined_df[col1], 
                        combined_df[f"{col2}_lag_{lag}"]
                    )
                    
                    # Add edge if correlation exceeds threshold
                    if abs(corr) > correlation_threshold:
                        dbn.add_edge(col2, col1, weight=abs(corr), lag=lag)
    
    return dbn
    
    def structural_equation_modeling(self):
        """
        Perform Structural Equation Modeling
        
        Returns:
            SEM model results and parameters
        """
        # Standardize data
        scaler = StandardScaler()
        scaled_data = scaler.fit_transform(self.data)
        
        # Use PyMC3 for SEM
        with pm.Model() as sem_model:
            # Latent variables
            latent_vars = {}
            for i in range(self.num_series):
                latent_vars[f'latent_{i}'] = pm.Normal(f'latent_{i}', 0, 1)
            
            # Structural relationships
            observations = {}
            for i in range(self.num_series):
                observations[f'obs_{i}'] = pm.Normal(
                    f'obs_{i}', 
                    mu=sum(latent_vars.values()),  # Simple summed latent effect
                    sigma=1
                )
            
            # Observe the data
            pm.Normal('likelihood', mu=observations, observed=scaled_data)
        
        # Sample from the model
        with sem_model:
            trace = pm.sample(2000, return_inferencedata=True)
        
        return {
            'model': sem_model,
            'trace': trace,
            'summary': az.summary(trace)
        }
    
    def causal_discovery_dowhy(self, treatment_var: int = 0, outcome_var: int = 1):
        """
        Causal Discovery using DoWhy
        
        Args:
            treatment_var: Index of treatment variable
            outcome_var: Index of outcome variable
        
        Returns:
            Causal inference results
        """
        # Prepare data
        data = self.df.copy()
        
        # Create causal graph
        graph = nx.DiGraph()
        graph.add_edges_from([(f'X{treatment_var}', f'X{outcome_var}')])
        
        # DoWhy causal model
        causal_model = dowhy.CausalModel(
            data=data,
            treatment=f'X{treatment_var}',
            outcome=f'X{outcome_var}',
            graph=graph
        )
        
        # Identify causal effect
        identified_estimand = causal_model.identify_effect()
        
        # Estimate causal effect
        estimate = causal_model.estimate_effect(identified_estimand)
        
        return {
            'causal_model': causal_model,
            'identified_estimand': identified_estimand,
            'estimate': estimate
        }
    
    def dynamic_causal_modeling(self):
        """
        Dynamic Causal Modeling (DCM)
        
        Returns:
            DCM model results
        """
        # Basic DCM implementation using PyMC3
        with pm.Model() as dcm_model:
            # State transitions
            x = pm.GaussianRandomWalk('state', sigma=1, shape=self.num_series)
            
            # Observation model
            obs_sigma = pm.HalfNormal('obs_sigma', sigma=1)
            obs = pm.Normal('obs', mu=x, sigma=obs_sigma, observed=self.data)
        
        # Sample from the model
        with dcm_model:
            trace = pm.sample(2000, return_inferencedata=True)
        
        return {
            'model': dcm_model,
            'trace': trace,
            'summary': az.summary(trace)
        }
    
    def causality_test(self, 
                       method: str = 'granger', 
                       lag: Optional[int] = None, 
                       verbose: bool = False, 
                       **kwargs):
        """
        Extended causality test method
        
        Args:
            method: Causality test method
            lag: Number of lags
            verbose: Verbose output
            **kwargs: Additional method-specific parameters
        
        Returns:
            Causality test results
        """
        # Existing methods plus new ones
        methods = {
            'granger': self.granger_causality,
            'dbn': self.dynamic_bayesian_network,
            'sem': self.structural_equation_modeling,
            'dowhy': self.causal_discovery_dowhy,
            'dcm': self.dynamic_causal_modeling
        }
        
        if method.lower() not in methods:
            raise ValueError(f"Method {method} not supported. Choose from {list(methods.keys())}")
        
        return methods[method.lower()](**kwargs)
